------------------------------------------------------------------------
-- Copyright Notice:
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory,
-- which is operated for the National Science Foundation by the
-- Association of Universities for Research in Astronomy under
-- Cooperative Agreement No. AST-0132798.
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CLK_SERVICES
-- MODULE       :   REGISTER CONTROL
-- VERSION      :   2.21
------------------------------------------------------------------------
-- FILE NAME    :   Clk_RegisterControl_Ver220.vhd
-- AUTHOR         :   Peter Moore
-- DATE               :   03/23/2012
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :  Registers to hold CLK Services module control parameters
--                         :  writen by the PAN or other modules across the WB Bus.
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :

------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	     AUTHOR  VERSION  MODIFICATIONS
-- 02/04/2010   PCM     1.02   INITIAL VERSION
-- 10/06/2010    PCM     2.01   Modifications made to the LMK3000 control module - This source has updated version number only
-- 12/17/2010    PCM     2.02   Changed way that clock conditioner load is initiated. Used a real one sysclk wide pulse rather than half the wishbone signalling
-- 03/05/2012   PCM     2.20   Version value only .. This allows a firmware recovery point to be set at system version 2.11 
-- 03/23/2012   PCM     2.21   Fixed LMK3000 config serial stream problem .. it was dropping bits occassionally  
------------------------------------------------------------------------
-- PORT DEFINITIONS :

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

entity Clk_RegisterControl_Ver221 is
	Port (
-- WishBone bus signals
		WbSysClk 				: in   std_logic;
		WbReset 					: in   std_logic;
		WbWriteEnable 			: in   std_logic;
		WbAddress 				: in   std_logic_vector(15 downto 0);
		WbDataIn 				: in   std_logic_vector(31 downto 0);
		WbWrtAck 				: out  std_logic;
		WbErrAck 				: out  std_logic;
		WbBusy 					: out  std_logic;
		WbDataOut 				: out  std_logic_vector(31 downto 0);
      WbStatOut            : out  std_logic_vector(7 downto 0);
      WbStatIn             : in   std_logic_vector(31 downto 0);
-- Module clock and reset
      reset_o              : out  std_logic;   -- Module reset pulse from wishbone bus
      Sysclk_o             : out  std_logic;
-- Clock control
      DheIsSlave_o         : out  std_logic;   -- Designates this DHE as a slave when true.
      DheSlaveClkSwitchEn_o: out  std_logic;   -- Enables the automatic clock switchover if a slave has en external clock input.
      DheSlaveClkMode_i    : in   std_logic;   -- If DHE is a slave and  is using external clock this signal is true 
-- Sync port clock control
      SyncClkSelect_o      : out  std_logic_vector(3 downto 0);
      SyncOutEqualization_o: out  std_logic_vector(1 downto 0);
      SyncInEqualization_o : out  std_logic_vector(1 downto 0);
      ClkCfgData_io        : inout std_logic_vector(31 downto 0);  -- Addressable portion of config regs
      ClkCfgAddr_o         : out  std_logic_vector(3 downto 0); -- Address to access regs in clock control module
      ClkCfgWrt_o          : out  std_logic;   -- Used to write data values into the registers
      clkCfgLoad_o         : out  std_logic;  -- Triggers a re-configuration of the clock conditioner using the user settable registers
-- Local status data
      LockDetect_i         : in   std_logic
 	);

end Clk_RegisterControl_Ver221;

architecture Behavioral of Clk_RegisterControl_Ver221 is

-- Firmware module code revision value
   constant CODE_IDENT   : std_logic_vector(15 downto 0) := x"00DD"; -- Current firmware version is 2.20
   constant MODULE_IDENT : std_logic_vector(15 downto 0) := x"00D0"; -- CLK Module identity code is 208
-- Misc bits --
   signal AddrErr : std_logic;      -- Indicates an addressing mismatch in this module i.e. bus master is trying to write to a non existant register

-- Decoding Vectors
   signal CS16K_0, CS16K_1, CS16K_2, CS16K_3 : std_logic;
   signal CS4K_0, CS4K_1, CS4K_2, CS4K_3 : std_logic;
   signal CS1K_0, CS1K_1, CS1K_2, CS1K_3 : std_logic;
   signal CS256_0, CS256_1, CS256_2, CS256_3 : std_logic;
   signal CS64_0, CS64_1, CS64_2, CS64_3 : std_logic;
   signal CS16_0, CS16_1, CS16_2, CS16_3 : std_logic;
   signal CS_0, CS_1, CS_2, CS_3, CS_4, CS_5, CS_6, CS_7 : std_logic;
   signal CS_8, CS_9, CS_A, CS_B, CS_C, CS_D, CS_E, CS_F : std_logic;

-- Misc registers and selects
   signal CodeId_cs, ModuleId_cs : std_logic;
	signal Address : std_logic_vector(15 downto 0);
	signal Data_in : std_logic_vector(31 downto 0);
   signal ClkServModInStatus_cs, ClkServModOutStatus_cs : std_logic;
   signal LocalStatus_reg : std_logic_vector(31 downto 0);
   
-- Address select signals for the Application Interface module
   signal DheSlaveClkSwitchEn_cs, DheSlaveClkMode_cs : std_logic;
   signal SyncClkSelect_cs, SyncOutEqualization_cs, SyncInEqualization_cs : std_logic;
   signal clkCfgLoad_cs, ClkCfgWrt_cs : std_logic;
   signal clkCfgLoad_reg : std_logic;
   signal SyncOutEqualization_reg, SyncInEqualization_reg : std_logic_vector(1 downto 0); 
   signal DheSlaveClkSwitchEn_reg     : std_logic;
   signal SyncClkSelect_reg           : std_logic_vector(3 downto 0);
   signal DheIsSlave                  : std_logic;
   
begin
------------------------------------------------------------------------
-- Combinational and Static logic assignments
------------------------------------------------------------------------
	Address <= WbAddress;
	Data_in <= WbDataIn;
	WbBusy <= '0';									-- Not used
	WbWrtAck <= WbWriteEnable when (AddrErr = '0' and WbWriteEnable = '1') else '0';
	WbErrAck <= AddrErr when (WbWriteEnable = '1') else '0';
   LocalStatus_reg <= ("0000000000000000000000000000000" & LockDetect_i);
   reset_o  <= WbReset;
   Sysclk_o <= WbSysClk;
   
-- Status bit assignments 
-- Output
   WbStatOut(0) <= LockDetect_i;
   WbStatOut(1) <= '0';
   WbStatOut(2) <= '0';
   WbStatOut(3) <= '0';
   WbStatOut(4) <= '0';
   WbStatOut(5) <= '0';
   WbStatOut(6) <= '0';
   WbStatOut(7) <= '0';
-- Input
   DheIsSlave   <= WbStatIn(7);
   
------------------------------------------------------------------------
-- Module output port assignments
------------------------------------------------------------------------
   DheIsSlave_o           <= DheIsSlave;
   DheSlaveClkSwitchEn_o  <= DheSlaveClkSwitchEn_reg; 
   SyncClkSelect_o        <= SyncClkSelect_reg;
   SyncOutEqualization_o  <= SyncOutEqualization_reg;
   SyncInEqualization_o   <= SyncInEqualization_reg;
   clkCfgLoad_o           <= clkCfgLoad_reg;
   ClkCfgWrt_o            <= ClkCfgWrt_cs and WbWriteEnable;
   ClkCfgAddr_o           <= Address(3 downto 0); -- Address to access regs in clock control module
   ClkCfgData_io          <= Data_in when (ClkCfgWrt_cs = '1' and WbWriteEnable = '1') else (others => 'Z');
    
------------------------------------------------------------------------
-- MODULE ADDRESS DECODE LOGIC. Enable the decoded address strobes as required.
------------------------------------------------------------------------
-- Decode 16K word Blocks
    CS16K_0 <= (not Address(15)) and (not Address(14));   -- 0x0000 => 0x3FFF
    CS16K_1 <= (not Address(15)) and Address(14);         -- 0x4000 => 0x7FFF
    CS16K_2 <= Address(15) and (not Address(14));         -- 0x8000 => 0xBFFF
    CS16K_3 <= Address(15) and Address(14);               -- 0xC000 => 0xFFFF

-- Decode 4K word Blocks
    CS4K_0 <= (not Address(13)) and (not Address(12));    --0x0000 => 0x0FFF, 0x4000 => 0x4FFF, 0x8000 => 0x8FFF, 0xC000 => 0xCFFF
    CS4K_1 <= (not Address(13)) and Address(12);          --0x1000 => 0x1FFF, 0x5000 => 0x5FFF, 0x9000 => 0x9FFF, 0xD000 => 0xDFFF
    CS4K_2 <= Address(13) and (not Address(12));          --0x2000 => 0x2FFF, 0x6000 => 0x6FFF, 0xA000 => 0xAFFF, 0xE000 => 0xEFFF
    CS4K_3 <= Address(13) and Address(12);                -- 0x3000 => 0x3FFF, 0x7000 => 0x7FFF, 0xB000 => 0xBFFF, 0xF000 => 0xFFFF

-- Decode 1K word Blocks
    CS1K_0 <= (not Address(11)) and (not Address(10));   -- 0x_000 => 0x_3FF
    CS1K_1 <= (not Address(11)) and Address(10);         -- 0x_400 => 0x_7FF
    CS1K_2 <= Address(11) and (not Address(10));         -- 0x_800 => 0x_BFF
    CS1K_3 <= Address(11) and Address(10);               -- 0x_C00 => 0x_FFF

-- Decode 256 word Blocks
    CS256_0 <= (not Address(9)) and (not Address(8));    --0x_000 => 0x_0FF, 0x_400 => 0x_4FF, 0x_800 => 0x_8FF, 0x_C00 => 0x_CFF
    CS256_1 <= (not Address(9)) and Address(8);          --0x_100 => 0x_1FF, 0x_500 => 0x_5FF, 0x_900 => 0x_9FF, 0x_D00 => 0x_DFF
    CS256_2 <= Address(9) and (not Address(8));          --0x_200 => 0x_2FF, 0x_600 => 0x_6FF, 0x_A00 => 0x_AFF, 0x_E00 => 0x_EFF
    CS256_3 <= Address(9) and Address(8);                -- 0x_300 => 0x_3FF, 0x_700 => 0x_7FF, 0x_B00 => 0x_BFF, 0x_F00 => 0x_FFF

-- Decode 64 word Blocks
    CS64_0 <= (not Address(7)) and (not Address(6));   -- 0x__00 => 0x__3F
    CS64_1 <= (not Address(7)) and Address(6);         -- 0x__40 => 0x__7F
    CS64_2 <= Address(7) and (not Address(6));         -- 0x__80 => 0x__BF
    CS64_3 <= Address(7) and Address(6);               -- 0x__C0 => 0x__FF

 -- Decode 16 word Blocks
    CS16_0 <= (not Address(5)) and (not Address(4));    --0x__00 => 0x__0F, 0x__40 => 0x__4F, 0x__80 => 0x__8F, 0x__C0 => 0x__CF
    CS16_1 <= (not Address(5)) and Address(4);          --0x__10 => 0x__1F, 0x__50 => 0x__5F, 0x__90 => 0x__9F, 0x__D0 => 0x__DF
    CS16_2 <= Address(5) and (not Address(4));          --0x__20 => 0x__2F, 0x__60 => 0x__6F, 0x__A0 => 0x__AF, 0x__E0 => 0x__EF
    CS16_3 <= Address(5) and Address(4);                -- 0x__30 => 0x__3F, 0x__70 => 0x__7F, 0x__B0 => 0x__BF, 0x__F0 => 0x__FF

-- Decode single word address lines
    CS_0 <= (not Address(3)) and (not Address(2)) and (not Address(1)) and (not Address(0));
    CS_1 <= (not Address(3)) and (not Address(2)) and (not Address(1)) and Address(0);
    CS_2 <= (not Address(3)) and (not Address(2)) and Address(1) and (not Address(0));
    CS_3 <= (not Address(3)) and (not Address(2)) and Address(1) and Address(0);
    CS_4 <= (not Address(3)) and Address(2) and (not Address(1)) and (not Address(0));
    CS_5 <= (not Address(3)) and Address(2) and (not Address(1)) and Address(0);
    CS_6 <= (not Address(3)) and Address(2) and Address(1) and (not Address(0));
    CS_7 <= (not Address(3)) and Address(2) and Address(1) and Address(0);
    CS_8 <= Address(3) and (not Address(2)) and (not Address(1)) and (not Address(0));
    CS_9 <= Address(3) and (not Address(2)) and (not Address(1)) and Address(0);
    CS_A <= Address(3) and (not Address(2)) and Address(1) and (not Address(0));
    CS_B <= Address(3) and (not Address(2)) and Address(1) and Address(0);
    CS_C <= Address(3) and Address(2) and (not Address(1)) and (not Address(0));
    CS_D <= Address(3) and Address(2) and (not Address(1)) and Address(0);
    CS_E <= Address(3) and Address(2) and Address(1) and (not Address(0));
    CS_F <= Address(3) and Address(2) and Address(1) and Address(0);

----------------------------------------------------------------------------------------------------
-- REGISTER ADDRESS DECODE LOGIC. Use the decoded address strobes to define a particular address.
----------------------------------------------------------------------------------------------------
-- From TRNT-AD-08-0003-Rev1.1
-- Each register description line will be a comment in the FPGA source code file of the form below. 
-- A comment  line  may be included in the file as a column header line.  
-- The column header comment line will look like this:
--
--##AD: EPAdr: AtrName: ArraySz: #Vals: TypCde: BdAdr: RegAdr: AdrIncr: Size: SetMthd: RdMthd: SigTyp: DspGrp: dfltVal: Units: Cmnt
--
---#AD is the tag indication for a normal register description line
-- EPAdr  This field will contain the starting word address (0-4095) in the EEPROM of the configuration information for this attribute. Acceptable values are an address starting at 0xFFF down to 0x000 in hex or decimal format or the words AUTO ,UNDCD or NONE. The AUTO / UNDCD words will indicate to assimilate that the location of the configuration data block for this attribute should be assigned automatically. The NONE word indicates to assimilate that no EEPROM words should be assigned to this attribute.
-- AtrName  The name used/defined in the FPGA. This is the DHE hardware name of the register/function.
-- ArraySz  the number of Registers in an array of registers with a common name and purpose. This can always be 1 treating each register in an array as a single register. The PAN already always unwraps arrays of registers to single registers.
-- #Vals  gives the size in words of the values needed to describe the behavior of the register. The values occupy EEPAdr  EEPAdr+(#Values-1)) in the EEPROM. These words contain the values described by TypCde below.
-- TypCde (Type Codes)
--       A  A single integer value entry representing information about the EEPROM structure or provenance; or
--             a single word engineering value stored as a 32 BIT binary value to be loaded automatically by the FPGA startup code; 
--       B  Two(if read only) or four(if read/write) integer values that describe a bit field register. Laid out as >Width, Mask, Max, Min< The Bit field is always assumed to be located in bits 0  width
--       C  Two(if read only) or four(if read/write) Floating point values that describe a value register. Laid out as >Slope,Intercept,Max,Min<
--       D  a single floating point value used in a software constant
--       E  a Four/Two word entry representing a checksum value least significant word first.
-- MdlName  the text name for the board/module containing the Register eg, CCD1, LCB
-- RegAddr: the 16 bit address of the register in the module address space of the first register in the group. The PAN will treat BdAddr and RegAddress: fields as a single address by summing the two fields with DheRegAddress = ((BdAddr<<16)&RegAddr+AdrIncr() ) The csv file will place a !BDNAME:0x04 line at the beginning of the file and describe the address as ,BDNAME:0234, in the register address field.
-- AdrIncr  A code indicating how addresses increment in arrays of registers. Three values are recognized at this time.
--       0, 1  indicate the addresses increment by 1 for each successive index i.e.. 1 2 3 4 5, etc.
--       2 - indicates the addresses increment by shifting the lower #inArray bits by one bit to the left. i.e.  0x1101, 0x1102, 0x1104, 0x1108, 0x1110, etc.
-- Size  The width in bits of the register, usually 32. Reasonable values are 1, 8, 12, 16, 24, 32.
-- SetMthd, RdMthd  The method used to write/read the register. Three methods are known:
--       NONE  The attribute cannot be accessed for writing or for reading.
--       SMPL  The register is accessed with a simple write or read command.
--       RDMW  The register is a bit field that requires a read, masking the bits that are unchanged and a simple write to set the register. On read the value returned is the value of the bits requested.
-- SigTyp  This gives the type of signal for connection purposes. The fpConfig routine will use this to limit the connection type of detector signals for this attribute. fpConfig will recognize the following signal types. Note: These types refer to the connections to the detector or Dewar and are not used for internal or logical function attributes. These attributes should be given the NONE, LGCL or SFTWR signal type.
--       DPWR  a detector power signal  a voltage usually connected to a fixed bias level.
--       DCLK  detector clock signal  a control clock for the detector. These break down into an UpperRail, a LowerRail and a logical level. Connecting any one of the three signals to a detector pin automatically connects the other two signals.
--       AGND  a detector analog ground signal  
--       OGND  other detector ground signals  
--       VOUT  a detector video output. This breaks out into the video signal and an offset voltage for that signal.
--       BIAS - a detector bias voltage.
--       DVLT  a detector voltage not otherwise classified (e.g. logical controls as in OTAs. or hardware gain controls).
--       NONE  an attribute that does not connect to the detector and for which no or only a write/read test exists. 
--       CMND  an attribute that when written to results in the start of some process in the DHE,  eg. afe1Reset, sysReset
--       TELE  an attribute that reads back the value of some telemetry voltage.
--       CNFG  a configuration Attribute  eg. Afe1TpGrp2Cfg or CcdAmpCfg
--       CNST  attributes that contain constants eg. CodeID, AFE1SerNum etcz
--       LGCL - an attribute that is a logical connection to the detector (i.e. rowCount, Number of outputs, pipeline enables, etc.)
--       TEMP  a temperature control signal.
--       SFTWR  connects to Dewar only through other values 
-- DspGrp  a number representing the display group that includes this attribute. Equivalent to the MEC or BORG page number to be used for this attributes display.
-- DfltVal  the value assigned to the attribute register when the fpga is loaded on power-up or when a hard reset occurs.
-- Units - Engineering units used to describe the attribute
-- Cmnt  a comment describing the usage/purpose/function of this register.
-----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
-- ATTRIBUTE DESCRIPTION:
-- #MD: CLK: 0x80:    --This is the Wishbone module select address of the Clock service module registers

-- #AD: NONE: ClkCodeId: 1: 1: A:  CLK: 0xFFFF: 0: 16:  NONE: SMPL: CNST: 10: 2.02: Version: Firmware revision code as MM.mm^MOD^ ^100.0^0.0^0.0^65535.0
   CodeId_cs <= CS16K_3 and CS4K_3 and CS1K_3 and CS256_3 and CS64_3 and CS16_3 and CS_F;

-- #AD: NONE:  ClkModuleId: 1: 1: A: CLK: 0xFFFE: 0: 16: NONE:  SMPL: CNST: 10: 208.0: Ident: Module identification code as an integer^MOD^ ^1.0^0.0^0.0^4294967295.0
-- #AD: NONE: ClkResetCmd: 1: 1: A:  CLK: 0xFFFE: 0: 0:  SMPL: NONE: CMND: 10: :ResetCmnd:Triggers a CLK module reset^MOD^ ^1.0^0.0^0.0^0.0
   ModuleId_cs <= CS16K_3 and CS4K_3 and CS1K_3 and CS256_3 and CS64_3 and CS16_3 and CS_E;

-- #AD: NONE: ClkModInStatus: 1: 1: A: CLK: 0xFFFD: 0: 32:  NONE: SMPL: NONE: 10: 0: Boolean: Status input from System.^MOD^ ^1.0^0.0^0.0^4294967295.0
   ClkServModInStatus_cs <= CS16K_3 and CS4K_3 and CS1K_3 and CS256_3 and CS64_3 and CS16_3 and CS_D;

-- #AD: NONE: ClkModOutStatus: 1:  1: A: CLK: 0xFFFC: 0:  32:  NONE:  SMPL: NONE: 10: 0: Boolean: Status of Clk Control module as transmitted to system.^MOD^ ^1.0^0.0^0.0^4294967295.0
   ClkServModOutStatus_cs <= CS16K_3 and CS4K_3 and CS1K_3 and CS256_3 and CS64_3 and CS16_3 and CS_C;

-- #AD: NONE: LoadClkCfg: 1: 1: A: CLK: 0x0015: 0: 1: SMPL: SMPL:  CMND: 10: 0: :Forces the re-loading of the clock conditioner constants - read shows the lock detect bit state^MOD^ ^1.0^0.0^0.0^1.0
   clkCfgLoad_cs  <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_0 and CS16_1 and CS_5;
  
-- #AD: NONE: ClkCfgRegs: 16: 1: A: CLK: 0x0070: 0: 32: SMPL: SMPL:  CNFG: 10: : Values: Clock conditioner configuration registers.^MOD^ ^1.0^0.0^0.0^4294967295.0
   ClkCfgWrt_cs <=  CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_1 and CS16_3;

-- #AD: NONE: SlaveClkXferEn: 1: 1: A: CLK: 0x0081: 0: 1: SMPL: SMPL:  CNFG: 10: 0: Boolean: Configure the DHE to switch to the SYNC clk when it is available and in configures as a Slave device^MOD^ ^1.0^0.0^0.0^1.0
   DheSlaveClkSwitchEn_cs <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_2 and CS16_0 and CS_1;
   
-- #AD: NONE: SlaveClkMode: 1: 1: A: CLK: 0x0082: 0: 1: NONE: SMPL:  NONE: 10: 0: Boolean: If DHE is a slave and is using external clock this signal is true^MOD^ ^1.0^0.0^0.0^1.0
   DheSlaveClkMode_cs <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_2 and CS16_0 and CS_2;

-- #AD: NONE: SyncClkSelect: 1: 1: A: CLK: 0x0083: 0: 4: SMPL: SMPL:  CNFG: 10: 0.0: Select: Selects and enables the source for the master SYNC clock output^MOD^ ^1.0^0.0^0.0^15.0
   SyncClkSelect_cs <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_2 and CS16_0 and CS_3;
   
-- #AD: NONE: SyncOutEqualization: 1: 1: A: CLK: 0x0084: 0: 2: SMPL: SMPL:  CNFG: 10: 0.0: Value: Selects the degree of pre-emphasis applied to the SYNC output signals^MOD^ ^1.0^0.0^0.0^3.0
   SyncOutEqualization_cs <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_2 and CS16_0 and CS_4; 
   
-- #AD: NONE: SyncInEqualization: 1: 1: A: CLK: 0x0085: 0: 2: SMPL: SMPL:  CNFG: 10: 0.0: Value: Selects the degree of de-emphasis applied to the SYNC input signals^MOD^ ^1.0^0.0^0.0^3.0
   SyncInEqualization_cs <= CS16K_0 and CS4K_0 and CS1K_0 and CS256_0 and CS64_2 and CS16_0 and CS_5;

----------------------------------------------------------------------------------------------------
-- REGISTER DATA LATCHING LOGIC. Use the decoded address CS strobes to enable a register write.
----------------------------------------------------------------------------------------------------
   LcbCtrlRegLatchProc : process (WbSysClk)
      begin
         if (rising_edge(WbSysClk)) then
            clkCfgLoad_reg <= '0';                -- Cancel any clock reload strobes from last clock cycle
            if (WbReset = '1') then               -- Presets (default values)
               DheSlaveClkSwitchEn_reg <= '0';    -- Disable clock hunting   
               SyncClkSelect_reg <= "0000";       -- Err .. I have forgotten what this does
               SyncOutEqualization_reg <= "00";   -- Minimum equilization
               SyncInEqualization_reg <= "00";    -- Minimum equilization
            elsif (WbWriteEnable = '1') then
               if (clkCfgLoad_cs = '1') then
                  clkCfgLoad_reg <= '1';
               end if;
               if (DheSlaveClkSwitchEn_cs = '1') then
                  DheSlaveClkSwitchEn_reg <= Data_in(0);
               end if;
               if (SyncClkSelect_cs = '1') then
                  SyncClkSelect_reg <= Data_in(3 downto 0);
               end if;
               if (SyncOutEqualization_cs = '1') then
                  SyncOutEqualization_reg <= Data_in(1 downto 0);
               end if;
                if (SyncInEqualization_cs = '1') then
                  SyncInEqualization_reg <= Data_in(1 downto 0);
               end if;
            end if;
         end if;
      end process;

--------------------------------------------------------------------------------------------------
-- Write addressing error catcher
--------------------------------------------------------------------------------------------------
   AddrErr <= not (DheSlaveClkSwitchEn_cs or SyncClkSelect_cs or SyncOutEqualization_cs or SyncInEqualization_cs or
                   clkCfgLoad_cs or ClkCfgWrt_cs);

----------------------------------------------------------------------------------------------------
-- REGISTER READ LOGIC. Use the **_CS strobes to multiplex the registers (or module input ports)
-- to the data_out port.
----------------------------------------------------------------------------------------------------

-- Add future registers here as they are generated.
-- Although the hierarchy here has an inherent priority, there should only be one address CS strobe
-- true on any given cycle, therefore there is no priority as to where the registers should go in this list.

   WbDataOut <=("0000000000000000000000000000000" & DheSlaveClkSwitchEn_reg)  when (DheSlaveClkSwitchEn_cs = '1') else
               ("0000000000000000000000000000000" & DheSlaveClkMode_i)        when (DheSlaveClkMode_cs = '1') else               
               ("0000000000000000000000000000" & SyncClkSelect_reg)           when (SyncClkSelect_cs = '1') else
               ("000000000000000000000000000000" & SyncOutEqualization_reg)   when (SyncOutEqualization_cs = '1') else               
               ("000000000000000000000000000000" & SyncInEqualization_reg)    when (SyncInEqualization_cs = '1') else 
               ("0000000000000000000000000000000" & LockDetect_i)             when (clkCfgLoad_cs = '1') else 
               ClkCfgData_io                                                  when (ClkCfgWrt_cs = '1') else              
               WbStatIn              								                  when (ClkServModInStatus_cs = '1') else
               LocalStatus_reg              								            when (ClkServModOutStatus_cs = '1') else
               ("0000000000000000" & MODULE_IDENT)					            	when (ModuleId_cs = '1') else
               ("0000000000000000" & CODE_IDENT)              						when (CodeId_cs = '1') else
               x"42424242"; --- Code word for bad bus read address :-)

end Behavioral;

