-----------------------------------------------------------------------
-- Copyright Notice:
-- (c) Copyright 2004 - 2008, National Optical Astronomy Observatory,
-- which is operated for the National Science Foundation by the
-- Association of Universities for Research in Astronomy under
-- Cooperative Agreement No. AST-0132798.
------------------------------------------------------------------------
-- PROJECT      :   MONSOON-TORRENT
-- BOARD        :   LOCAL CONTROL BOARD
-- DEVICE ID    :   VIRTEX 5LX
-- PARENT MODULE:   CFG_SERVICES
-- MODULE       :   CLOCK CONTROL
-- VERSION      :   2.21
------------------------------------------------------------------------
-- FILE NAME    :   LMK03000_Control_V221.vhd
-- AUTHOR        :   Peter Moore
-- DATE              :   03/23/2012
-- COMPANY      :   NOAO
-- PLATFORM     :   Xilinx ISE 10.1.03
------------------------------------------------------------------------
-- DESCRIPTION  :  Controls the primary clock source for the DHE
-- 
------------------------------------------------------------------------
-- IMPLEMENTATION NOTES :
-- Set the reset bit to ensure the device is in the default state
-- Clock divider registers bit format: [31:17]=0, [16]=EN, [15:8]=DIV, [7:4]=DLY, [3:0]=ADDR
	-- clk_reg(0)  <= x"00030800";		-- SYNC_CLK	- div by 16 = 39.84375 MHz
	-- clk_reg(1)  <= x"00030301";		-- LCLK - div by 6 = 106.25 MHz
	-- clk_reg(2)  <= x"00000102";		-- Not Used
	-- clk_reg(3)  <= x"00030803";		-- TCLK - div by 16 = 39.84375 MHz
	-- clk_reg(4)  <= x"00030804";		-- MCLK - div by 16 = 39.84375 MHz
	-- clk_reg(5)  <= x"00030405";		-- DCLK - div by 8 = 79.6875 MHz
	-- clk_reg(6)  <= x"00030206";		-- PCLK - div by 4 = 159.375 MHz
	-- clk_reg(7)  <= x"00000107";		-- Not Used
-- Clock configuration registers:
--
-- New value with external loop filter capacitors and primary clock running at 53.125 MHz
	-- clk_reg(9)  <= x"0082000B";	   -- R11 -  Div4 = 0
	-- clk_reg(10) <= x"028D495D";		-- R13  - Input f. = 53, R3/R4 = 10K, C3 = 100pf, C4 = 110pf.
	-- clk_reg(11) <= x"0830040E";		-- R14  -  GlobalEnableClk = 1, DigitalLockDetHigh, PLL_R_Div = 4
	-- clk_reg(12) <= x"4800300F";		-- R15  -  CP gain = 4, VCO_DIV = 2, PLL_N_Div = 48

-- Proposed new value with external loop filter capacitors and primary clock running at 106.25 MHz
	-- clk_reg(9)  <= x"0082000B";	   -- R11 -  Div4 = 0
	-- clk_reg(10) <= x"029A895D";		-- R13  - Input f. = 106, R3/R4 = 10K, C3 = 100pf, C4 = 110pf.
	-- clk_reg(11) <= x"0830080E";		-- R14  -  GlobalEnableClk = 1, DigitalLockDetHigh, PLL_R_Div = 8
	-- clk_reg(12) <= x"4800300F";		-- R15  -  CP gain = 4, VCO_DIV = 2, PLL_N_Div = 48
--
   -- Original values with no loop filter and primary clock running at 106.25 MHz
	-- clk_reg(8)  <= x"0082000B";	   -- R11  -  Div4 = 0
	-- clk_reg(9) <= x"029A800D";		-- R13  - Input f. = 106, R3/R4 = 600, C3 = 0pf, C4 = 110pf.
	-- clk_reg(10) <= x"0850080E";		-- R14  -  GlobalEnableClk = 1, AnalogDetHigh, PLL_R_Div = 8
	-- clk_reg(11) <= x"0800300F";		-- R15  -  CP gain = 1, VCO_DIV = 2, PLL_N_Div = 48
   
   -- This setup requires a low pass filter connected to pin 28 of the LMK03000 device with the following component values (see LMK03000 data sheet for schematic)
   -- C1 = 5.6nf
   -- C2 = 7.5nf
   -- R2 = 3.6K Ohms
------------------------------------------------------------------------
-- REVISION HISTORY :
-- DATE	     AUTHOR  VERSION  MODIFICATIONS
-- 06/02/2009   DGS   1.00        INITIAL VERSION
-- 06/22/2009   PCM    1.01        based on original design and added two banks of config registers, triggerable from command line, and slave switching logic
-- 02/22/2010    PCM    1.02       Changed the module name from ClkServices_V101 to LMK03000_Control_V102 to reflect reality and incorporate into the higher level module called CLK_Services_Ver102.
--                                                     Changed control constants to run at 53.125 MHz
-- 10/06/2010     PCM    2.01      Changed LMK3000 loader FSM to now start sending the LMK3000 register control words from reg address zero (not one). This fixes the bug that prevented SYNC CLK to be generated. 
-- 12/17/2010    PCM    2.02      Changes to try and find out why the initial system clock conditioner load fails sometimes ... Produces just one CLK_CNTRL_CLK pulse before dying ... More  news later 
-- 03/23/2012   PCM   2.21      Found that the config serial stream was occassionally dropping bits that confused the clock conditioner. This version fixes that problem.
------------------------------------------------------------------------
-- PORT DEFINITIONS :

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity LMK03000_Control_V221 is
   Port ( 
      SysClk         : in  std_logic;
      Reset          : in  std_logic;
      CLK_CTRL_CLK   : out std_logic;  -- LMK03000 serial command clock
      CLK_CTRL_DATA  : out std_logic;  -- LMK03000 serial command data stream
      CLK_CTRL_CS_N  : out std_logic;  -- LMK03000 chip select 
      CLK_CTRL_SYNC  : out std_logic;  -- LMK03000 signal to sync clocks to positive edge
      CLK_CTRL_GOE   : out std_logic;  -- LMK03000 signal to globally enable clocks
      CLK_LOCK_DET   : in  std_logic;
      MCLK_SEL_N     : out std_logic;  -- powers down the local LCB oscillator
      SYNC_CLK_SNS   : in  std_logic;  -- Clock signal from master - used to detect an active external master clock
      DheIsSlave_i   : in  std_logic;  -- Signal toindicate that the DHE is designated as a slave 
      ClkSwitchEn_i  : in  std_logic;  -- enables the DHE to automatically select the external clock if present
      DheClkMode_o   : out std_logic;   -- Indicates the actual clock source being used - true indicates external clock is used.
      ClkCfgData_io  : inout std_logic_vector(31 downto 0);  -- Addressable portion of config regs
      ClkCfgAddr_i   : in  std_logic_vector(3 downto 0); -- Address to access regs
      ClkCfgWrt_i    : in  std_logic;   -- Used to write data values into the registers
      clkCfgLoad_i   : in  std_logic;   -- Triggers a re-configuration of the clock conditioner using the user settable registers (clk_reg(31:16))
      ClkLockOut_o   : out std_logic;   -- Goes false when the conditioner is being programmed - to hold off the GPT clock generator from locking too early
      ClkLockDet_o   : out std_logic    -- The clock conditioner lock detect signal
      );
end LMK03000_Control_V221;

architecture Behavioral of LMK03000_Control_V221 is

	constant SCLK_DIV : std_logic_vector(3 downto 0)     :="0011";				      -- setting +1 determines half serial clock period in sysclk cycles. 3 = approx. 6.75 MHz @ 53.125MHz SysClk 
	constant SYNC_DLY : std_logic_vector(4 downto 0)     :="00111";					-- used for sync pulse duration = Approx. 130ns
	constant SETTLE_DLY : std_logic_vector(4 downto 0)   :="01111";				   -- used for clock settling duration = Approx. 260ns
 --  constant CLKCON_RST : std_logic_vector(31 downto 0) := x"80000000";		-- bit 31, register zero is reset for the conditioning device
   constant BOOT_DELAY : std_logic_vector(17 downto 0)  := "101011001101101000"; -- 18-bit delay for 10ms after boot before programming the clock synthesizer
   constant CLK_WORD_CNT : std_logic_vector(3 downto 0) := "1101";               -- Number of words from the rom to use to initialize clock conditioner
   

-- LMK03000 constants...
-- CLK_IN frequency = 106.25 MHz / 2 = 53.125 MHz
-- VCO R-DIV value = 4
-- VCO N-DIV value = 48
-- VCO-DIV value = 2
-- VCO-DIV-4 value = 0 (i.e. no division on phsae detector)
-- Therefore VCO frequency (VCO-FREQ) =  CLK_IN * VCO-DIV * N-DIV / R-DIV = 53.125 * 2 * 48 / 4 = 1275 MHz 
-- CLK_OUT frequencies will be VCO-FREQ / VCO-DIV  / CLK_OUT_DIV  
-- CLK0 = SYNC_OUT_CLK = CLK_OUT_DIV value = 16 = 39.84375 MHz
-- CLK1 = LCLK = CLK_OUT_DIV value = 6 = 106.25 MHz
-- CLK2 = Disabled = CLK_OUT_DIV value = 16 = Not Used 
-- CLK3 = TCLK = CLK_OUT_DIV value = 16 = 39.84375 MHz
-- CLK4 = MCLK = CLK_OUT_DIV value = 6 = 106.25 MHz
-- CLK5 = DCLK = CLK_OUT_DIV value = 8 = 79.6875 MHz
-- CLK6 = PCLK = CLK_OUT_DIV value = 4 = 159.375 MHz
-- CLK7 = Disabled = CLK_OUT_DIV value = 16 = Not used 
   
   
	type reg_type is array (31 downto 0) of std_logic_vector (31 downto 0);
-- For 53.125 MHz   
   signal clk_reg : reg_type := (x"00000000", x"00000000", x"00000000", x"4800300F", 
                                 x"0830040E", x"028D495D", x"0082000B", x"00000807", 
                                 x"00030206", x"00030405", x"00030304", x"00030803",
                                 x"00000802", x"00030301", x"00030800", x"80000000",
                                 x"00000000", x"00000000", x"00000000", x"4800300F", 
                                 x"0830040E", x"028D495D", x"0082000B", x"00000807", 
                                 x"00030206", x"00030405", x"00030304", x"00030803",
                                 x"00000802", x"00030301", x"00030800", x"80000000");
                                 
	signal ClkRomData, buff_reg : std_logic_vector(31 downto 0);
	signal sclk, clk_en, clk_cs_n, clk_data, clk_sync, clk_cfg_trig, clk_cfg_trig_ack : std_logic;
   signal cfg_status : std_logic := '0';    -- On power up or boot, do the clock configuration once. 
	signal sclk_ctr, word_cnt : std_logic_vector(3 downto 0);
   signal sync_ctr : std_logic_vector(4 downto 0); 
   signal addr_index, array_ptr : std_logic_vector(4 downto 0);     -- Clock Reg Array pointer 
	signal bit_cntr : std_logic_vector(4 downto 0);
	type statetype is (cfg_idle, cfg_wrt, cfg_dly, cfg_done, cfg_sync, cfg_settle);
	signal cfg_state : statetype;
   signal ClkSelect, SlaveClkActive : std_logic;    -- This bit is set true when running on the slave clock source
   signal SlvSnsGateCntr, SlvClkCntr : std_logic_vector(3 downto 0); 
   signal BootDelayCntr : std_logic_vector(17 downto 0);
   signal ClkStat : std_logic;

begin

   ClkStat <= CLK_LOCK_DET and cfg_status;
   ClkLockOut_o <= ClkStat when (cfg_state = cfg_idle) else '0';
   ClkLockDet_o <= CLK_LOCK_DET;

	CLK_CTRL_GOE <= '1';
	CLK_CTRL_DATA <= clk_data;
	CLK_CTRL_SYNC <= clk_sync;
	CLK_CTRL_CS_N <= clk_cs_n;
   
   
     
-- Interface to the PAN via the register control module   
   array_ptr <= '1' & ClkCfgAddr_i when (cfg_state = cfg_idle) else addr_index;

   WrtClkCfgRegsProc : process(SysClk)
   begin
      if (rising_edge(SysClk)) then
         if (ClkCfgWrt_i = '1') then
            clk_reg(conv_integer(array_ptr)) <= ClkCfgData_io;
         end if;
      end if;
   end process;
   
   ClkRomdata <= clk_reg(conv_integer(array_ptr));

   ClkCfgData_io <=  ClkRomData when (ClkCfgWrt_i = '0') else (others => 'Z');
   
-- Clock change over logic
   ClkChangeProc : process(SysClk, Reset, ClkSwitchEn_i)
   begin
      if (Reset = '1' or ClkSwitchEn_i = '0') then
            ClkSelect <= '0';
            SlvSnsGateCntr <= (others => '0');
      elsif (rising_edge(SysClk)) then
         if (DheIsSlave_i = '1') then     -- We are enabled to switch to slave clock when we can
            SlvSnsGateCntr <= SlvSnsGateCntr + 1;     -- Increment the gate counter to detect the slave clock
            if (SlaveClkActive = '1') then
               ClkSelect <= '1';          -- All conditions are met.
            end if;
         end if;
      end if;
   end process;

   MCLK_SEL_N   <= '1'; -- Locked down to trace initialization problem ... not ClkSelect;
   DheClkMode_o <= ClkSelect;

   SlaveClkSnsProc : process(SYNC_CLK_SNS, SlvSnsGateCntr)
   begin
      if (SlvSnsGateCntr = "0000") then
         SlvClkCntr <= (others => '0');
         if (SlvClkCntr >= "1101") then
            SlaveClkActive <= '1';     -- We have accumulated active clock cycles .. the slave clock appears to be alive
         end if;
      elsif (rising_edge(SYNC_CLK_SNS)) then
            SlvClkCntr <= SlvClkCntr + 1;
      end if;
   end process;
   
------------------------------------------------------------------------
-- SCLK Process: Generate a serial clock for data transfer to LMK03000
------------------------------------------------------------------------
	SerialClockProc : process (SysClk)
	begin
		if (rising_edge (SysClk)) then
         if (Reset = '1') then
			   sclk_ctr <= (others => '0');
			   sclk <= '0';
			elsif (sclk_ctr < SCLK_DIV) then
				sclk_ctr <= sclk_ctr + 1;
			else
				sclk_ctr <= (others => '0');
				sclk <= not sclk;
			end if;
		end if;
	end process;

	CLK_CTRL_CLK <= sclk and clk_en;
--	CLK_CTRL_CLK <= sclk when (clk_en = '1') else '0';

   ClockConfigTriggerProc : process (SysClk)
	begin
		if (rising_edge (SysClk)) then
         if (Reset = '1' or clk_cfg_trig_ack = '1') then
            clk_cfg_trig <= '0';
         elsif (clkCfgLoad_i = '1') then
            clk_cfg_trig <= '1';
         end if;
      end if;
   end process;
   
------------------------------------------------------------------------
-- Clock Configuration Process: Set up the LMK03000 registers
------------------------------------------------------------------------
	ConfigureClockProc : process (Reset, sclk)
	begin
		if (Reset = '1') then
			clk_en <= '0';
			clk_cs_n <= '0';
			clk_data <= '0';
			clk_sync <= '1';
         clk_cfg_trig_ack <= '0';
			cfg_state <= cfg_idle;
		elsif (falling_edge (sclk)) then
			case cfg_state is
				when cfg_idle =>
               word_cnt <= (others => '0');        -- Get ready to count the output registers .. a total of 12 registers + reset word are loaded 
               bit_cntr <= (others => '0');
               if (cfg_status = '0') then
                  if (BootDelayCntr < BOOT_DELAY) then
                     BootDelayCntr <= BootDelayCntr + 1;
                     cfg_state <= cfg_idle;        -- wait for 10ms after a cold boot before programming the clock generator
                  else
			            clk_cs_n <= '1';
                     clk_en <= '1';
                     addr_index <= "00000";         -- Load from the untouchables
                     cfg_state <= cfg_done;
                  end if;                     
					elsif (clk_cfg_trig = '1') then		-- we are commanded so reconfigure the clock conditioner
						clk_en <= '1';
			         clk_cs_n <= '1';
                  addr_index <= "10000";           -- Load data from the settable registers  if not reset triggered
                  cfg_status <= '0';               -- set the lock indication false to the system while we reboot the clock controller
						cfg_state <= cfg_done;
					else
			         clk_cs_n <= '0';
                  clk_cfg_trig_ack <= '0';
						cfg_state <= cfg_idle;
					end if;
				when cfg_wrt =>
				   clk_cs_n <= '0';                             -- Enable the data xfer frame
					clk_data <= buff_reg(31);                    -- Output the msb ...
					buff_reg <= buff_reg(30 downto 0) & "0";		-- shift data left
					if (bit_cntr = "11111") then     -- 31, last bit
				      sync_ctr <= (others => '0');
						cfg_state <= cfg_dly;
					else
						bit_cntr <= bit_cntr + 1;
						cfg_state <= cfg_wrt;
					end if;
            when cfg_dly =>
					clk_cs_n <= '1';                             -- Set end of frame marker 
					if (sync_ctr < SYNC_DLY) then                -- Delay for to get data into lmk03000 register
						sync_ctr <= sync_ctr + 1;
						cfg_state <= cfg_dly;
					else
--						addr_index <= addr_index + 1;                -- increment the rom store word address -- Moved 03/23/2012 to allow more time for rom to respond.
						cfg_state <= cfg_done;                       -- check to see if we are finished here
					end if;
				when cfg_done =>
					if (word_cnt < CLK_WORD_CNT) then
						buff_reg <= ClkRomdata;                   -- Load the next word
                  word_cnt <= word_cnt + 1;                 -- Increment the word counter - require n words to configure the device
						addr_index <= addr_index + 1;             -- increment the rom store word address to next channel
						bit_cntr <= (others => '0');
						cfg_state <= cfg_wrt;
					else
                  clk_en <= '0';                      -- Switch off the clock
 					   clk_data <= '0';                    -- output quiescent state                
                  sync_ctr <= (others => '0');        -- Get ready to output sync pulse
						cfg_state <= cfg_sync;
					end if;
				when cfg_sync =>
					if (sync_ctr < SYNC_DLY) then
						clk_sync <= '0';
						sync_ctr <= sync_ctr + 1;
						cfg_state <= cfg_sync;
					else
						clk_sync <= '1';
                  sync_ctr <= (others => '0');
						cfg_state <= cfg_settle;
					end if;
				when cfg_settle =>
					if (sync_ctr < SETTLE_DLY) then     -- Wait fo rthe clocks to settle before going live.
						sync_ctr <= sync_ctr + 1;
						cfg_state <= cfg_settle;
					else
						cfg_status <= '1';               -- Lock out any retrigger of the boot load sequence
					   clk_data <= '0';
                  clk_cfg_trig_ack <= '1';         -- Acknowledge to clear the trigger flag 
						cfg_state <= cfg_Idle;
               end if;
				when others =>
					cfg_state <= cfg_idle;
			end case;
		end if;
	end process;
	
end Behavioral;

